#
# Importe le maillage sur l'image du monde
# Sélection de l'objectif
# Construction de la table de transition
# Construction de la table des valeurs
# Mise en place de l'algorithme avec les eq. de Belleman

from imutils.video import VideoStream
import argparse
import imutils
from pyzbar.pyzbar import decode
import openfoodfacts
import requests
import json
import subprocess

import meshio
import numpy as np
import pandas as pd
import cv2
import os
import time
from time import sleep

#from random import randint
import RPi.GPIO as GPIO

global mouseX, mouseY
global cible_selected
global image_cache
global image_originale
global img
global distance_cible_courante
global distance_cible_pecedente

#Pour developement :
import random
testCode = 3228470065412

cible_selected = False

ETAT_DEPART = 86  # Etat de départ
ETAT_CIBLE = -1  # Etat cible
EPSILON = 0.85  # Politique "greddy" (0.0 : non greddy / 1.0 :  greddy)
ALPHA = 0.1  # Taux d'apprentissage
ACTIONS = ['A1', 'A2', 'A3']  # Liste des actions

RECOMPENSE = 50.0
RECOMPENSE_MUR = -0.1
RECOMPENSE_NON_CIBLE = -0.1
PROBA_MAX = 1.0
PROBA_MIN = (1.0 - PROBA_MAX) / 2


#####################################
# Callback d'un clique sur la souris
#
def get_mouse_click(event, x, y, flags, params):
    global mouseX, mouseY
    global cible_selected
    if event == cv2.EVENT_LBUTTONUP:
        mouseX, mouseY = x, y
        cible_selected = True


###############################################################
# Fonction permettant de trouver l'état ciblé par une coordonée
# Entrées:  array : Coordonnées de l'ensemble des états
#           values : Coordonnées de la cible
# Sortie:   idx : index de l'état le plus proche de la cible
#
def find_nearest(array, value):
    # Classement des éléments les plus proches en norme
    norms = np.linalg.norm(array, axis=1)
    dif = np.abs(norms - np.linalg.norm(value))
    idx_norm = dif.argsort()

    # Classement des éléments les plus proches en coordonnées X
    dif = np.abs(array[:, 0] - value[0])
    idx_x = dif.argsort()

    # Retrouve le premier élement commun en X et en norme
    for value_norm in idx_norm:
        find = False
        for i in range(0, 5):
            if value_norm == idx_x[i]:
                idx = idx_x[i]
                find = True
                break
        if find == True:
            break
    # Récupère le premier élement commun
    return idx


###########################################################
# Fonction permettant de récupérer les 3 états voisins
# autour d'un état dont l'index est donné
# Enrée:    idx: index de l'état cible
# Sortie:   etats_voisins[] : Liste des états voisins
def GetEtatsVoisins(idx):
    etats_voisins = [-1, -1, -1]
    liste = [0, 1, 2, 0]
    for i in range(3):
        noeud1 = noeuds_surface[idx, liste[i]]
        noeud2 = noeuds_surface[idx, liste[i + 1]]

        for j in range(noeuds_surface.shape[0]):
            if j == idx:
                continue
            if (noeud1 in noeuds_surface[j]) and (noeud2 in noeuds_surface[j]):
                etats_voisins[i] = j
                centerX = int(triangles_surface[j, 3, 0] / (1000 * res))
                centerY = int(triangles_surface[j, 3, 1] / (1000 * res))
                break
    return etats_voisins


########################################################
# Fonction permettant de déterminer l'action à effectuer
# à partir de l'état courant en fonction de la table
# des valeurs
# Entrée :      ETAT_COURANT    :   Etat courant
#               tables_valeurs  :   table des valeurs
# Sortie:       A:                  Action (0,1 ou 2)
def get_action(ETAT_COURANT, table_valeurs):
    actions_etat_courant = table_valeurs[ETAT_COURANT, :]  # Récupère les valeurs des actions sur l'état courant

    # Si on tire au sort une valeur > EPSILON
    # ou si la table des valeurs est vierge
    # Alors on choisit une action au hasard parmi celles disponibles
#    rng = np.random.default_rng()
#    val = rng.uniform()

#    if (val > EPSILON) or ((actions_etat_courant == 0).all()):
    if 1 == 2:
        A = rng.choice([0, 1, 2])

    # Sinon, on suit la meilleure action à prendre
    # dans l'état courant
    else:  # act greedy
        A = np.argmax(np.array(actions_etat_courant))
    return A


###########################################
# Détemrine l'état suivant en fonction de l'état courant et de l'action choisie
# à l'aide de la tables des probabilités de transitions
#
# Entrée :      ETAT_COURANT :          Etat courant
#               A :                     Action (0,1,2)
#               table_transition:       Table des probabilités des transitions : (nbr_etats,3,nbr_etats)
#               table_recompenses:      Table des récompenses : (nbr_etats,3,nbr_etats
# Sortie :      ETAT_SUIVANT:           Etat suivant
#               R :                     Récompense

def get_action_suivante(ETAT_COURANT, A, table_transition, force=False):
    global distance_cible_courante
    global distance_cible_pecedente
    etat_deja_visite = True

    table_probas = table_transition[ETAT_COURANT, A, :]['proba_transition']  # tenseur au format (nbr_etats,)
    rng = np.random.default_rng()

    if force == True:
        val = np.argmax(table_probas)
    else:
        # Tire au sort l'état suivant en utilisant les probabilitées contenues dans la table des transitions
        val = rng.choice([i for i in range(3)], 1, p=np.squeeze(table_probas, 1)).item()

    ETAT_SUIVANT = table_transition[ETAT_COURANT, A, val]['index_etat_suivant'].item()

    return ETAT_SUIVANT


######################################################
# Fonction permettant d'afficher le chemin parcouru
# sur l'image
# Entrée :  ETAT_COURANT    :   Index de l'état courant
#           ETAT_SUIVAN     :   Index de l'état suivant
#           img             :   Image sur laquelle dessiner
def AfficheChemin(ETAT_COURANT, ETAT_SUIVANT):
    global image_cache
    global img

    img = image_cache.copy()
    pos_ETAT_COURANT = triangles_surface[ETAT_COURANT, 3, :]
    pos_ETAT_SUIVANT = triangles_surface[ETAT_SUIVANT, 3, :]

    x1 = int(pos_ETAT_COURANT[0] / (1000 * res))
    x2 = int(pos_ETAT_SUIVANT[0] / (1000 * res))
    y1 = int(pos_ETAT_COURANT[1] / (1000 * res))
    y2 = int(pos_ETAT_SUIVANT[1] / (1000 * res))

    centerX = int(triangles_surface[ETAT_SUIVANT, 3, 0] / (1000 * res))
    centerY = int(triangles_surface[ETAT_SUIVANT, 3, 1] / (1000 * res))
    img = cv2.line(img, (x1, img.shape[0] - y1), (x2, img.shape[0] - y2), color=(0x1C, 0x8E, 0x00), thickness=2)
    image_cache = img.copy()

    img = cv2.circle(img, (centerX, img.shape[0] - centerY), radius=2, color=(0, 255, 0), thickness=2)
    cv2.imshow('Image', img)
    cv2.waitKey(1)


# Fonction d'évaluation d'une politique
# Permet d'obtenir les valeurs des états ("value fonction")
# par itérations successives jusqu'à convergence
# vk+1(s) = f(vk(s)) => page 74 et 75
# Entrées :     pi : Politique à évaluer
#               table_transition : Table des transitions (nbr_etats, nbr_actions, etat_suivant) = (nbr_etats, 3, nbr_etats)
#               table_recompenses : Table des récompenses : (nbr_etats,3,nbr_etats)
def Evaluation_politique(pi, table_transition, GAMMA, THETA, V):
    # Initialise la valeur des états aléatoirement (ici à 0)
#    Vinit = V.copy()
#    V_ = Vinit.copy()
    ETATS_VISITES = []
    V_ = np.zeros(triangles_surface.shape[0], dtype=np.float32)

    while True:
        V = np.zeros(triangles_surface.shape[0], dtype=np.float64)      # Vk+1
#        V = Vinit.copy()
        for etat in range(triangles_surface.shape[0]):
            # Calcul Vk+1(etat)
            action = pi(etat)
            for etat_suivant in range(3):
                index_etat_suivant = table_transition[etat, action, np.int8(etat_suivant)]['index_etat_suivant'].item()
                prob = table_transition[etat, action, np.int8(etat_suivant)]['proba_transition'].item()
                R = table_transition[etat, action, np.int8(etat_suivant)]['recompense'].item()

                if index_etat_suivant in ETATS_VISITES:
                    R = RECOMPENSE_RETOUR_ARRIERE
                if index_etat_suivant != ETAT_CIBLE:
                    V[etat] = V[etat] + prob * (R + GAMMA * V_[np.int8(index_etat_suivant)])
                else:
                    V[etat] = V[etat] + prob * R
#                ETATS_VISITES.append(index_etat_suivant)

#        print("%.20f" % np.max(np.abs(V - V_)))
        if np.max(np.abs(V - V_)) < THETA:
            break
        V_ = V.copy()
    return V


# Fonction d'amélioration d'une politique
def Amelioration_politique(V, table_transition, GAMMA):
    ETATS_VISITES = []
    # Initialise table des valeurs/actions (ici à 0) Q(s,a)
    Q = np.zeros((triangles_surface.shape[0], 3), dtype=np.float32)

    for etat in range(triangles_surface.shape[0]):
        for action in range(3):
            for etat_suivant in range(3):
                index_etat_suivant = table_transition[etat, action, np.int8(etat_suivant)]['index_etat_suivant'].item()
                prob = table_transition[etat, action, np.int8(etat_suivant)]['proba_transition'].item()
                R = table_transition[etat, action, np.int8(etat_suivant)]['recompense'].item()

                if index_etat_suivant in ETATS_VISITES:
                    R = RECOMPENSE_RETOUR_ARRIERE
#                ETATS_VISITES.append(etat_suivant)

                if index_etat_suivant != ETAT_CIBLE:
                    Q[etat, action] = Q[etat, action] + prob * (R + GAMMA * V[index_etat_suivant])
                else:
                    Q[etat, action] = Q[etat, action] + prob * R

    def new_pi(s): return {s: a for s, a in enumerate(np.argmax(Q, axis=1))}[s]
    return new_pi

def IterationValeur(table_transition, GAMMA, THETA):
    ETATS_VISITES = []
    V = np.zeros(triangles_surface.shape[0], dtype=np.float32)     # Vk
    while True:
        Q = np.zeros((triangles_surface.shape[0], 3), dtype=np.float32)
        for etat in range(triangles_surface.shape[0]):
            for action in range(3):
                for etat_suivant in range(3):
                    index_etat_suivant = table_transition[etat, action, np.int8(etat_suivant)]['index_etat_suivant'].item()
                    prob = table_transition[etat, action, np.int8(etat_suivant)]['proba_transition'].item()
                    R = table_transition[etat, action, np.int8(etat_suivant)]['recompense'].item()

                    if index_etat_suivant != ETAT_CIBLE:
                        Q[etat, action] = Q[etat, action] + prob * (R + GAMMA * V[index_etat_suivant])
                    else:
                        Q[etat, action] = Q[etat, action] + prob * R

#        print(np.max(np.abs(V - np.max(Q, axis=1))))
        if np.max(np.abs(V - np.max(Q, axis=1))) < THETA:
            break
        V = np.max(Q, axis=1)

    def pi(s): return {s: a for s, a in enumerate(np.argmax(Q, axis=1))}[s]
    return V,pi


def ShowPolitique(pi, V, table_valeurs, force=False,max_step=200):
    Tableau_etats = [];
    # Affiche le chemin en suivant la politique Greddy
    ETAT_COURANT = ETAT_DEPART
    apprentissage_fini = False
    EPSILON = 1.0  # Politique "greddy"
    numero_step = 0

    image_cache = image_originale.copy()
    img = image_originale.copy()
    image_cache = img.copy()
    cv2.imshow('Image', img)
    cv2.waitKey(1)

    # Construction de la table des valeurs
    for s in range(triangles_surface.shape[0]):
        table_valeurs[s, :] = -1e10
        table_valeurs[s, pi(s)] = V[s]

    numero_step = 0
    while not apprentissage_fini:
        # Sélectionne une action à effectuer
        A = get_action(ETAT_COURANT, table_valeurs)  # A = 0,1 ou 2

        # Applique l'action sur l'état courant ETAT_COURANT
        # et récupère le nouvel état ETAT_SUIVANT
        # et la récompense R
        ETAT_SUIVANT = get_action_suivante(ETAT_COURANT, A, table_transition, force=force)
        Tableau_etats.append(ETAT_SUIVANT);

        # Trace le chemin à parcourir
        AfficheChemin(ETAT_COURANT, ETAT_SUIVANT)

        if ETAT_SUIVANT == ETAT_CIBLE or numero_step == max_step:
            apprentissage_fini = True

        # Passe à l'état suivant
        ETAT_COURANT = ETAT_SUIVANT
        numero_step = numero_step + 1
    return Tableau_etats

def detect(image): #Un seul argument correspondant à l'image ou à la vidéo dans laquelle on souhaite detecter le code barre
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY) # Conversion de l'image en nuances de gris

    #Dans chacun des axes X et Y on applique un gradient de Scharr représentant les directions de l'images 
    ddepth = cv2.cv.CV_32F if imutils.is_cv2() else cv2.CV_32F
    gradX = cv2.Sobel(gray, ddepth=ddepth, dx=1, dy=0, ksize=-1)
    gradY = cv2.Sobel(gray, ddepth=ddepth, dx=0, dy=1, ksize=-1)
    #Soustraction du gradient Y au gradient X

    gradient = cv2.subtract(gradX, gradY)
    gradient = cv2.convertScaleAbs(gradient)
    # Flou et seuil de l'image :
    blurred = cv2.blur(gradient, (9, 9))
    (_, thresh) = cv2.threshold(blurred, 225, 255, cv2.THRESH_BINARY)
    # Construction du noyau et floutage du reste
    
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (21, 7))
    closed = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel)
    # Application d'une série de claque
    closed = cv2.erode(closed, None, iterations=4)
    closed = cv2.dilate(closed, None, iterations=4)
    # Recherche du contour de l'image non floutée
    
    cnts = cv2.findContours(closed.copy(), cv2.RETR_EXTERNAL,
        cv2.CHAIN_APPROX_SIMPLE)
    cnts = imutils.grab_contours(cnts)
    # Si tout est flou (aucune bordure trouvée)
    if len(cnts) == 0:
        return None
    # Sinon on sort les contours du code barre
    # Et on retourne dans le sens du plus long bord vers le bas
    c = sorted(cnts, key=cv2.contourArea, reverse=True)[0]
    rect = cv2.minAreaRect(c)
    box = cv2.cv.BoxPoints(rect) if imutils.is_cv2() else cv2.boxPoints(rect)
    box = np.int0(box)
    # Retour de l'image correspondant au code barre
    return box

def say(something, voice='mb-fr4'):
    print("espeak -v {} {}".format(voice,something))
    subprocess.call(['espeak', '-v%s' % (voice), something])


########################### LANCEMENT Conversion magasin/mesh ##################################


res = 0.05  # Résolution (5cm/pixel)

img = cv2.imread('map.png', cv2.IMREAD_COLOR)

# Extraction des informations sur le maillage
# mesh.cells_dict['triangle'] : (948,3):  contient le numéro des noeuds de chaque triangle
# mesh.points : (448,3) : Contient les coordonnées des noeuds
mesh = meshio.read("map_to_mesh.msh")

# Création du tenseur des triangles contenant les corrdonées (X,Y,Z)
# format : (nombre_triangles,[noeud1,noeud2,noeud3n,centre],coordonnées_XYZ)
triangles = np.zeros((len(mesh.cells_dict['triangle']), 4, 3), np.float32)

# Création du tenseur des noeuds contenant le numéro des noeuds
# format : (nombre_triangles,3)
noeuds = np.zeros((len(mesh.cells_dict['triangle']), 3), np.float32)

for i in range(0, triangles.shape[0]):
    noeud1 = mesh.cells_dict['triangle'][i][0]
    noeud2 = mesh.cells_dict['triangle'][i][1]
    noeud3 = mesh.cells_dict['triangle'][i][2]

    noeuds[i, 0] = noeud1
    noeuds[i, 1] = noeud2
    noeuds[i, 2] = noeud3

    centerX = (mesh.points[noeud1][0] + mesh.points[noeud2][0] + mesh.points[noeud3][0]) / 3;
    centerY = (mesh.points[noeud1][1] + mesh.points[noeud2][1] + mesh.points[noeud3][1]) / 3;
    centerZ = (mesh.points[noeud1][2] + mesh.points[noeud2][2] + mesh.points[noeud3][2]) / 3;

    triangles[i, 0, :] = mesh.points[noeud1]
    triangles[i, 1, :] = mesh.points[noeud2]
    triangles[i, 2, :] = mesh.points[noeud3]
    triangles[i, 3, :] = (centerX, centerY, centerZ)

# Sélection des triangles et des noeuds en surface
# format : (nombre_triangles,[noeud1,noeud2,noeud3,centre],coordonnées_XYZ)
triangles_surface = np.squeeze(triangles[np.where(triangles[:, 3, 2] == 5), :, :], 0)
noeuds_surface = np.squeeze(noeuds[np.where(triangles[:, 3, 2] == 5), :], 0)

print(triangles_surface.shape[0])
print("triangles_surface.shape[0]") #A Suppr

# Enregistre les noeuds correspondants

# Affichage des triangles sur l'image avec les centres
for i in range(0, triangles_surface.shape[0]):
    x1 = int(triangles_surface[i, 0, 0] / (1000 * res))
    x2 = int(triangles_surface[i, 1, 0] / (1000 * res))
    x3 = int(triangles_surface[i, 2, 0] / (1000 * res))

    y1 = int(triangles_surface[i, 0, 1] / (1000 * res))
    y2 = int(triangles_surface[i, 1, 1] / (1000 * res))
    y3 = int(triangles_surface[i, 2, 1] / (1000 * res))

    centerX = int(triangles_surface[i, 3, 0] / (1000 * res))
    centerY = int(triangles_surface[i, 3, 1] / (1000 * res))

    img = cv2.line(img, (x1, img.shape[0] - y1), (x2, img.shape[0] - y2), color=(0, 0, 255), thickness=1)
    img = cv2.line(img, (x2, img.shape[0] - y2), (x3, img.shape[0] - y3), color=(0, 0, 255), thickness=1)
    img = cv2.line(img, (x3, img.shape[0] - y3), (x1, img.shape[0] - y1), color=(0, 0, 255), thickness=1)

    img = cv2.circle(img, (centerX, img.shape[0] - centerY), radius=1, color=(255, 0, 0), thickness=2)

cv2.imshow('Image', img)
cv2.setMouseCallback('Image', get_mouse_click)
cv2.imwrite("map_etats.png", img)

# Récupération de l'état objectif
#

####################### DEBUT PROGRAMME ####################################

ligne = 0

BoutonOUI = 3
GPIO.setmode(GPIO.BCM)

GPIO.setup(BoutonOUI, GPIO.IN)

#Ouverture de la liste de course
#Lecture produit ligne 1
#Si appui sur bouton "oui" alors chercher produit

Liste = open("ListeCourse.txt", 'r')

Produit = Liste.readlines()[ligne]

print("Le produit recherché est :")
print(Produit)
say("Le produit recherché est :"+ str(Produit))

ReponseOUI=GPIO.input(BoutonOUI)
print( "Attente de réponse ")
sleep(0.2)
print(ReponseOUI)
while ReponseOUI == 1:
    ReponseOUI=GPIO.input(BoutonOUI)
    print( "Attente de réponse ")
    sleep(0.2)
    print(ReponseOUI)
    
if ReponseOUI == 0:
    print("Recherche du produit dans le magasin")
    say("Nous recherchons l'accès à "+ str(Produit))
        # Retrouve le numéro de l'état
        
        #idx = find_nearest(triangles_surface[:, 3, 0:2], np.array([mouseX * 1000 * res, (img.shape[0] - mouseY) * 1000 * res]))
    placementMagasin = open("PositionProduit.txt", 'r')
    idx=int(placementMagasin.readlines()[ligne])
        
    ETAT_CIBLE = idx
    print(idx)

        # Entoure l'état cible sélectionné sur l'image
    centerX = int(triangles_surface[idx, 3, 0] / (1000 * res))
    centerY = int(triangles_surface[idx, 3, 1] / (1000 * res))
    img = cv2.circle(img, (centerX, img.shape[0] - centerY), radius=2, color=(0, 0, 255), thickness=3)
    cv2.imshow('Image', img)
    cv2.waitKey(1)
    image_cache = img.copy()
    image_originale = img.copy()

        ############################################################
        # Construction de la table des probabilités de transitions
        # et des récompenses
        # Format :  (nbr_etats, nbr_actions, nbr_etat_suivant,[idex_etat_suivant,proba_transition,recompense]) = (nbr_etats, 3, 3,[1,1,1])
        # actions : 'A1' : 0
        #           'A2' : 1
        #           'A3' : 2
        # Probabilités: 0% de chance d'aller sur un "état suivant" qui n'est pas un état voisin de l'état courant
        #               100% de chance d'aller sur un état suivant en suivant une action particulière :
        #                   Ex : si p(s',R|s,A1) = 1.0 alors => p(s',R|s,A2) = 0.0 et p(s',R|s,A3) = 0.0
        #               0% de chance d'aller sur un "état suivant" si son numéro est "-1" (c'est un mur)
        # Récompenses:  +1 si "état suivant" est la cible
        #               0  sinon
    print("Construction de la table des transitions")
    say("recherche de la trajectoire")
    structure_table = np.dtype(
        [('index_etat_suivant', np.int32), ('proba_transition', np.float32), ('recompense', np.float32)])
    table_transition = np.zeros((triangles_surface.shape[0], 3, 3, 1),
                                    dtype=structure_table)  # (nbr_etats,nbr_actions,nbr_etat_suivant,1)

    data = []
    columns = ['Etat', 'Action', 'Etat suivant', 'Proba', 'Récompense']

    for index_etat in range(0, triangles_surface.shape[0]):
        for action in range(0, 3):
            for next_etat in range(3):
                table_transition[index_etat, action, next_etat]['recompense'] = RECOMPENSE_NON_CIBLE

    for index_etat in range(0, triangles_surface.shape[0]):
        etats_voisins = GetEtatsVoisins(index_etat)
        if np.count_nonzero(np.asarray(etats_voisins) == -1) == 0:
            etat_le_plus_proche = np.abs(np.asarray(etats_voisins) - index_etat)
            etat_le_plus_proche = etats_voisins[np.argmin(etat_le_plus_proche)]
        else:
            etats_voisins_ = np.delete(etats_voisins, np.where(np.asarray(etats_voisins) == -1))
            etat_le_plus_proche = np.abs(np.asarray(etats_voisins_) - index_etat)
            etat_le_plus_proche = etats_voisins_[np.argmin(etat_le_plus_proche)]

            # 3 états voisins
            # p(etat_voisin[0],R|etat_courant,A1) = 1.0 ; p(etat_voisin[0],R|etat_courant,A2) = p(etat_voisin[0],R|etat_courant,A3) = 0.0
            # p(etat_voisin[1],R|etat_courant,A2) = 1.0 ; p(etat_voisin[1],R|etat_courant,A1) = p(etat_voisin[1],R|etat_courant,A3) = 0.0
            # p(etat_voisin[2],R|etat_courant,A3) = 1.0 ; p(etat_voisin[2],R|etat_courant,A1) = p(etat_voisin[2],R|etat_courant,A3) = 0.0
        if np.count_nonzero(np.asarray(etats_voisins) == -1) == 0:
            table_transition[index_etat, 0, 0]['index_etat_suivant'] = etats_voisins[0]
            table_transition[index_etat, 0, 1]['index_etat_suivant'] = etats_voisins[1]
            table_transition[index_etat, 0, 2]['index_etat_suivant'] = etats_voisins[2]
            table_transition[index_etat, 0, 0]['proba_transition'] = PROBA_MAX
            table_transition[index_etat, 0, 1]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 0, 2]['proba_transition'] = PROBA_MIN
            
            table_transition[index_etat, 1, 0]['index_etat_suivant'] = etats_voisins[0]
            table_transition[index_etat, 1, 1]['index_etat_suivant'] = etats_voisins[1]
            table_transition[index_etat, 1, 2]['index_etat_suivant'] = etats_voisins[2]
            table_transition[index_etat, 1, 0]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 1, 1]['proba_transition'] = PROBA_MAX
            table_transition[index_etat, 1, 2]['proba_transition'] = PROBA_MIN

            table_transition[index_etat, 2, 0]['index_etat_suivant'] = etats_voisins[0]
            table_transition[index_etat, 2, 1]['index_etat_suivant'] = etats_voisins[1]
            table_transition[index_etat, 2, 2]['index_etat_suivant'] = etats_voisins[2]
            table_transition[index_etat, 2, 0]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 2, 1]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 2, 2]['proba_transition'] = PROBA_MAX

            if etats_voisins[0] == ETAT_CIBLE:
                table_transition[index_etat, 0, 0]['recompense'] = RECOMPENSE
                table_transition[index_etat, 1, 0]['recompense'] = RECOMPENSE
                table_transition[index_etat, 2, 0]['recompense'] = RECOMPENSE
            elif etats_voisins[1] == ETAT_CIBLE:
                table_transition[index_etat, 0, 1]['recompense'] = RECOMPENSE
                table_transition[index_etat, 1, 1]['recompense'] = RECOMPENSE
                table_transition[index_etat, 2, 1]['recompense'] = RECOMPENSE
            elif etats_voisins[2] == ETAT_CIBLE:
                table_transition[index_etat, 0, 2]['recompense'] = RECOMPENSE
                table_transition[index_etat, 1, 2]['recompense'] = RECOMPENSE
                table_transition[index_etat, 2, 2]['recompense'] = RECOMPENSE

            # 2 états voisins
            # p(etat_voisin[0],R|etat_courant,A1) = 1.0 ; p(etat_voisin[0],R|etat_courant,A2) = p(etat_voisin[0],R|etat_courant,A3) = 0.0
            # p(etat_voisin[1],R|etat_courant,A2) = 1.0 ; p(etat_voisin[1],R|etat_courant,A1) = p(etat_voisin[1],R|etat_courant,A3) = 0.0
            # p(etat_voisin[2],R|etat_courant,A1) = p(etat_voisin[2],R|etat_courant,A2) = p(etat_voisin[2],R|etat_courant,A3) = 0.0
        elif np.count_nonzero(np.asarray(etats_voisins) == -1) == 1:
            table_transition[index_etat, 0, 0]['index_etat_suivant'] = etats_voisins_[0]
            table_transition[index_etat, 0, 1]['index_etat_suivant'] = etats_voisins_[1]
            table_transition[index_etat, 0, 2]['index_etat_suivant'] = index_etat
            table_transition[index_etat, 0, 0]['proba_transition'] = PROBA_MAX
            table_transition[index_etat, 0, 1]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 0, 2]['proba_transition'] = PROBA_MIN

            table_transition[index_etat, 1, 0]['index_etat_suivant'] = etats_voisins_[0]
            table_transition[index_etat, 1, 1]['index_etat_suivant'] = etats_voisins_[1]
            table_transition[index_etat, 1, 2]['index_etat_suivant'] = index_etat
            table_transition[index_etat, 1, 0]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 1, 1]['proba_transition'] = PROBA_MAX
            table_transition[index_etat, 1, 2]['proba_transition'] = PROBA_MIN

            table_transition[index_etat, 2, 0]['index_etat_suivant'] = etats_voisins_[0]
            table_transition[index_etat, 2, 1]['index_etat_suivant'] = etats_voisins_[1]
            table_transition[index_etat, 2, 2]['index_etat_suivant'] = index_etat
            table_transition[index_etat, 2, 0]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 2, 1]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 2, 2]['proba_transition'] = PROBA_MAX

            if etats_voisins_[0] == ETAT_CIBLE:
                table_transition[index_etat, 0, 0]['recompense'] = RECOMPENSE
                table_transition[index_etat, 1, 0]['recompense'] = RECOMPENSE
                table_transition[index_etat, 2, 0]['recompense'] = RECOMPENSE

            elif etats_voisins_[1] == ETAT_CIBLE:
                table_transition[index_etat, 0, 1]['recompense'] = RECOMPENSE
                table_transition[index_etat, 1, 1]['recompense'] = RECOMPENSE
                table_transition[index_etat, 2, 1]['recompense'] = RECOMPENSE

            table_transition[index_etat, 0, 2]['recompense'] = RECOMPENSE_MUR
            table_transition[index_etat, 1, 2]['recompense'] = RECOMPENSE_MUR
            table_transition[index_etat, 2, 2]['recompense'] = RECOMPENSE_MUR

            # 1 état voisin
            # p(etat_voisin[0],R|etat_courant,A1) = 1.0 ; p(etat_voisin[0],R|etat_courant,A2) = p(etat_voisin[0],R|etat_courant,A3) = 0.0
            # p(etat_voisin[1],R|etat_courant,A1) = p(etat_voisin[1],R|etat_courant,A2) = p(etat_voisin[1],R|etat_courant,A3) = 0.0
            # p(etat_voisin[2],R|etat_courant,A1) = p(etat_voisin[2],R|etat_courant,A2) = p(etat_voisin[2],R|etat_courant,A3) = 0.0
        elif np.count_nonzero(np.asarray(etats_voisins) == -1) == 2:
            table_transition[index_etat, 0, 0]['index_etat_suivant'] = etats_voisins_[0]
            table_transition[index_etat, 0, 1]['index_etat_suivant'] = index_etat
            table_transition[index_etat, 0, 2]['index_etat_suivant'] = index_etat
            table_transition[index_etat, 0, 0]['proba_transition'] = PROBA_MAX
            table_transition[index_etat, 0, 1]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 0, 2]['proba_transition'] = PROBA_MIN

            table_transition[index_etat, 1, 0]['index_etat_suivant'] = etats_voisins_[0]
            table_transition[index_etat, 1, 1]['index_etat_suivant'] = index_etat
            table_transition[index_etat, 1, 2]['index_etat_suivant'] = index_etat
            table_transition[index_etat, 1, 0]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 1, 1]['proba_transition'] = PROBA_MAX
            table_transition[index_etat, 1, 2]['proba_transition'] = PROBA_MIN

            table_transition[index_etat, 2, 0]['index_etat_suivant'] = etats_voisins_[0]
            table_transition[index_etat, 2, 1]['index_etat_suivant'] = index_etat
            table_transition[index_etat, 2, 2]['index_etat_suivant'] = index_etat
            table_transition[index_etat, 2, 0]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 2, 1]['proba_transition'] = PROBA_MIN
            table_transition[index_etat, 2, 2]['proba_transition'] = PROBA_MAX

            if etats_voisins_[0] == ETAT_CIBLE:
                table_transition[index_etat, 0, 0]['recompense'] = RECOMPENSE
                table_transition[index_etat, 1, 0]['recompense'] = RECOMPENSE
                table_transition[index_etat, 2, 0]['recompense'] = RECOMPENSE

            table_transition[index_etat, 0, 1]['recompense'] = RECOMPENSE_MUR
            table_transition[index_etat, 0, 2]['recompense'] = RECOMPENSE_MUR
            table_transition[index_etat, 1, 1]['recompense'] = RECOMPENSE_MUR
            table_transition[index_etat, 1, 2]['recompense'] = RECOMPENSE_MUR
            table_transition[index_etat, 2, 1]['recompense'] = RECOMPENSE_MUR
            table_transition[index_etat, 2, 2]['recompense'] = RECOMPENSE_MUR

    for index_etat in range(0, triangles_surface.shape[0]):
        for action in range(0, 3):
            for next_etat in range(0, 3):
                data.append([index_etat, action, table_transition[index_etat, action, next_etat]['index_etat_suivant'],
                             table_transition[index_etat, action, next_etat]['proba_transition'],
                             table_transition[index_etat, action, next_etat]['recompense']])

     # Affiche la table
    columns = ['Etat', 'Action', 'Etat suivant', 'Proba', 'Récompense']
    df_table = pd.DataFrame(data=data, columns=columns)
    print(df_table)

        ###########################################################
        # Apprentissage : Algorithme de type "Policy iteration"
        #
    table_valeurs = np.zeros((triangles_surface.shape[0], 3), np.float32)  # (nbr_etats,nbr_actions)
    df_table_valeurs = pd.DataFrame(data=table_valeurs, columns=['A1', 'A2', 'A3'])

    THETA = 1e-10  # Seuil de convergence de l'évaluation de la politique
    GAMMA = 0.9  # Discount factor

    mode_visualisation = True

    rng = np.random.default_rng()
    actions = rng.choice(3, triangles_surface.shape[0])
    def pi(s): return {s: a for s, a in enumerate(actions)}[s]


    V = np.zeros(triangles_surface.shape[0], dtype=np.float64)  # Vk

    image_cache = image_originale.copy()
    img = image_originale.copy()
    image_cache = img.copy()

    while True:
        V, pi = IterationValeur(table_transition, GAMMA, THETA)

#        print([pi(i) for i in range(triangles_surface.shape[0])])
            
            #for i in range(triangles_surface.shape[0]):
                #print(pi(i))
                #print("i")
        img = image_originale.copy()
        image_cache = image_originale.copy()
        break

        # Affichage de la politique "pi" avec les valeurs optimale


    tableau_etats = ShowPolitique(pi, V, table_valeurs, force=True)
    
# format : (nombre_triangles,[noeud1,noeud2,noeud3,centre],coordonnées_XYZ)
#triangles_surface = np.squeeze(triangles[np.where(triangles[:, 3, 2] == 5), :, :], 0)
    etat_1 = ETAT_DEPART
    for etat_plus1 in tableau_etats:
        pos_courante = triangles_surface[etat_1,3,:]
        pos_plus1 = triangles_surface[etat_plus1,3,:]
        vecteur_directeur = [pos_plus1[0] - pos_courante[0],pos_plus1[1] - pos_courante[1],0];
        if vecteur_directeur[0] > 0 and vecteur_directeur[1] > 0:
            print("droite [%.1f]" %(vecteur_directeur[0]/1000));
            print("haut [%.1f]\n" %(vecteur_directeur[1]/1000));
        elif vecteur_directeur[0] > 0 and vecteur_directeur[1] < 0:
            print("droite [%.1f]" %(vecteur_directeur[0]/1000));
            print("bas [%.1f]\n" %(vecteur_directeur[1]/1000));
        elif vecteur_directeur[0] < 0 and vecteur_directeur[1] < 0:
            print("gauche [%.1f]" %(vecteur_directeur[0]/1000));
            print("bas [%.1f]\n" %(vecteur_directeur[1]/1000));
        elif vecteur_directeur[0] < 0 and vecteur_directeur[1] > 0:
            print("gauche [%.1f]" %(vecteur_directeur[0]/1000));
            print("haut [%.1f]\n" %(vecteur_directeur[1]/1000));
        elif vecteur_directeur[0] > 0 and vecteur_directeur[1] == 0:
            print("droite [%.1f]\n" %(vecteur_directeur[0]/1000));
        elif vecteur_directeur[0] < 0 and vecteur_directeur[1] == 0:
            print("gauche [%.1f]\n" %(vecteur_directeur[0]/1000));
        elif vecteur_directeur[1] > 0 and vecteur_directeur[0] == 0:
            print("haut [%.1f]\n" %(vecteur_directeur[0]/1000));
        elif vecteur_directeur[1] < 0 and vecteur_directeur[0] == 0:
            print("bas [%.1f]\n" %(vecteur_directeur[0]/1000));

        pos_1 = pos_plus1
            
            
        
        
    
    
    ligne = ligne + 1 # On passe à la ligne suivante dans la liste
    
    Liste.close()
    
    ap = argparse.ArgumentParser()
    ap.add_argument("-v", "--video",
    help="path to the (optional) video file")
    args = vars(ap.parse_args())
    # if the video path was not supplied, grab the reference to the
    # camera
    if not args.get("video", False):
        vs = VideoStream(src=0).start()
        time.sleep(2.0)
    # otherwise, load the video
    else:
        vs = cv2.VideoCapture(args["video"])
     #keep looping over the frames
    while True:
        # grab the current frame and then handle if the frame is returned
        # from either the 'VideoCapture' or 'VideoStream' object,
        # respectively
        frame = vs.read()
        frame = frame[1] if args.get("video", False) else frame
     
        # check to see if we have reached the end of the
        # video
        if frame is None:
            break
        #testCode = random.randrange(0, 1000000000)
        
        # detect the barcode in the image
        box = detect(frame)
        # if a barcode was found, draw a bounding box on the frame
        if box is not None:
            cv2.drawContours(frame, [box], -1, (0, 255, 0), 2)
        # show the frame and record if the user presses a key
        cv2.imshow("Frame", frame)
        key = cv2.waitKey(1) & 0xFF
        # if the 'q' key is pressed, stop the loop
        #Deplacer ca en dessous de frame
        ###########################################
        CBnumber = testCode
        CB = str(CBnumber)
        r = requests.get("https://world.openfoodfacts.org/api/v0/product/" + CB + ".json")
        #ingredients = openfoodfacts.facets.get_ingredients("GET/product/3017620422003")
        result = json.loads(r.text)
        
        lectureDebut = str(result).find('keywords')
        lectureFin = str(result).find('added_countries_tags')
        Donnee = str(result)[lectureDebut:lectureFin]
        File = open('Vocal', 'w')
        File.write(Donnee)
        File = open('Vocal', 'r')
        Lecture = File.read()
        say(Lecture)
        
        print(result)
        
        ############################################
        if key == ord("q"):
            break
    # if we are not using a video file, stop the video file stream
    if not args.get("video", False):
        vs.stop()
    # otherwise, release the camera pointer
    else:
        vs.release()
    # close all windows
    
        



    
########### Pour Developpement ###########
    
#print("Cliquer sur l'état objectif")
#cible_selected = False
#while (1):
#    cv2.waitKey(1)
#    if cible_selected is True:
#        break


###################CAMERA/Code Barre################



# Affichage de la politique "pi" avec les valeurs optimale
#ShowPolitique(pi, V, table_valeurs, force=True)
cv2.waitKey(0)
cv2.destroyAllWindows()




